"""Extremely simple time-out queues for channels

These do *not* attempt to store/manage state, they *only*
queue up updates for clients who are temporarily unable to
retrieve them (e.g. because of network failures or regular
timeout/restart of connections).
"""
import bisect,weakref,time,logging,collections
log = logging.getLogger( 'channels')

class Channel( object ):
    """A particular (active) channel being listened to by a user

    Messages should only be added to the channel in their final form,
    i.e. after all updates to the db...
    """
    messages = collections.deque()
    watchers = []
    STALE_PERIOD = 5*60
    @classmethod
    def get_existing( cls, key ):
        """Retrieve any existing channel for the given key"""
        return cls( key )
    @classmethod
    def get_channel( cls, key, state=None ):
        """Find channel of the given key"""
        return cls( key )
    def __init__( self, key, state=None ):
        """Initialize this channel with the given key"""
        self.key = str(key)
        self.state = state
    def add_watcher( self, watcher ):
        """Add a single watcher to the set of watchers"""
        self.watchers.append( watcher )
    def get_state( self ):
        return self.state
    def send( self, message ):
        """Send message to this channel"""
        message['channel'] = self.key
        message['ts'] = time.time()
        self.messages.append( message)
        # Note: not thread safe!
        new_watchers = []
        for watcher in self.watchers:
            watcher( message )
            if not watcher.invalid:
                new_watchers.append( watcher )
        self.__class__.watchers = new_watchers
        return message['ts']
    @classmethod
    def _remove_old( cls ):
        stale = time.time() - cls.STALE_PERIOD
        try:
            message = cls.messages.popleft()
            while message['ts'] < stale:
                message = cls.messages.popleft()
            cls.messages.appendleft( message )
        except IndexError, err:
            pass
    @classmethod
    def _for_key( cls, key, since ):
        result = []
        for message in reversed(cls.messages):
            if message['ts'] < since:
                if message['channel'] == key:
                    result.append( message )
            else:
                break
        return reversed(result)

    def since( self, ts ):
        """Return all messages since given ts

        This is to be called when something wants to check for updates,
        we'll have up to timeout seconds worth of previous messages and
        can (reasonably) efficiently check for any new messages...
        """
        self._remove_old()
        return self._for_key( self.key, ts )
