// Copyright 2009 FriendFeed
//
// Licensed under the Apache License, Version 2.0 (the "License"); you may
// not use this file except in compliance with the License. You may obtain
// a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

// This is the Tornado Long-Polling client-side code and support stuff
// broken out into a separate file and made slightly more reuse-friendly
// than the original chat demo.

function getCookie(name) {
    var r = document.cookie.match("\\b" + name + "=([^;]*)\\b");
    return r ? r[1] : undefined;
}

if (!Array.prototype.push) {
    Array.prototype.push = function( x ) {
        this[this.length] = x;
    };
}
if (!Array.prototype.forEach) {
    Array.prototype.forEach = function( f ) {
        var result = [];
        for (var i=0;i<this.length;i++) {
            result.push( f( this[i] ));
        }
    };
}

/* Provide browser-specific information, currently we don't have a good algo for determining
which browser supports which transports.  Chrome and IE seem to fall over on streaming
xmlhttprequests, so they fall back to long-polling. */

$.support.xmlhttpreq_stream = (!($.browser.msie));

$.longPoll = function( config ) {
    /* Create a new long-polling object to interact with a Tornado web-server

    longPoll( {} ) or longPoll( 'url' )

    Long-polling objects are actually multi-mode objects, they can be long-polling,
    quick-polling or streaming.  Setting pollInterval to a non-null value will create
    a quick-polling object.  Setting stream to true will create a streaming-long-polling
    object.

    returns a longPoll object

    The longPoll object has these attributes which can be passed in within the config
    parameter:

        url -- the URL from which to request updates...

        method -- the HTTP method to use to request updates (default 'POST')

        stream -- if true, request streaming response from the server...

        callbacks -- an array to which you can append code which needs to be called
                          on each successful load.  function( json_result, updater ) {} is called for each
                          successful poll.  default []

        baseArguments -- an object (dictionary) of parameters which will be passed to the
                                    polling query.  This can be specified e.g. to restrict the query to
                                    a subset of values or what have you.

        calcArguments -- if provided, a function to calculate arguments for each call instead, these
                                   will then have baseArguments

        failureCallbacks -- an array to which to append code which needs to be called
                                     on every failed post.  Note: this is not edge-triggering, that is,
                                     it is called every time we try-and-fail, not just when "going offline"
                                     default []

        pollInterval -- if set, will pause between polls for this period (milliseconds), used if
                              you want to do non-long-polling (i.e. regular polling) requests

        errorSleepTime -- an internal value telling us how long to wait after a failure to
                                    attempt a reconnect

        live -- flag controlled by stop/restart

    methods/entry-points:

        poll() -- function which starts polling the server

        stop() -- function which causes the polling to stop (eventually), cancels timeouts and
                     sets a flag to prevent the next poll from happening.  TODO: store the ajax
                     request object and cancel that too.

        restart() -- function which resets the poll-suppression flag and calls poll()

    */
    if (!window.console) { window.console = {}; }
    if (!window.console.log) { window.console.log = function() {}; }
    /* TODO: maybe switch to new/prototype base implementation (i.e. this)? */
    if (! config.url ) {
        config = { url: config };
    }
    var updater = {
        callbacks: [],
        failureCallbacks: [],
        method: 'POST',
        /* baseArguments can be modified by outside code to e.g. add cursors or other metadata */
        /* TODO: _xsrf support seems a bit... too specific... */
        baseArguments: {},
        calcArguments: null,
        errorSleepTime: 500,
        pollInterval: 0,
        stream: $.support.xmlhttpreq_stream,

        request: null,
        timer: null,

        pollCount: 0,
        bytesRead: 0,
        live: true,
        MESSAGE_SPLIT_TEXT: '""""""""""""""""""""""""""""',
        /* random values are so cool... */
        MAX_STREAM_DATA:  1300*1024
    };
    $.extend( updater, config );

    updater.abortCurrent = function() {
        /* abort any current in-process operations */
        if (updater.timer) {
            console.log( 'Clearing longPoll timer' );
            window.clearTimeout( updater.timer );
            updater.timer = null;
        }
        if (updater.request) {
            console.log( 'Cancelling longPoll request' );
            updater.request.abort();
            updater.request = null;
            updater.pollCount = 0;
        }
    };
    updater._onSuccess = function( result ) {
        updater.errorSleepTime = 500; /* reset as we're back to working */
        updater.timer = null;
        for (var i=0;i<updater.callbacks.length;i++) {
            try {
                updater.callbacks[i]( result, updater );
            } catch (e) {
                try {
                    console.log( "Failure during callback "+e+": "+updater.callbacks[i] );
                } catch (err) {
                    /* TODO: why does that fail... hrmph, likely a scope issue */
                }
            }
        }
    };
    updater.onSuccess = function( result ) {
        /* Long-polling version... */
        result = JSON.parse( result );
        updater.request = null;
        if (result.error) {
            /* request succeeded, but we got an error at the application level */
            console.log( 'JSON error reported ' + result.error );
            return updater.onError( result );
        } else {
            updater.pollCount -= 1;
            updater._onSuccess( result );
            updater.poll();
        }
    };
    updater.onSuccessNoWait = function( result ) {
        result = JSON.parse( result );
        updater.request = null;
        if (result.error) {
            /* request succeeded, but we got an error at the application level */
            console.log( 'JSON error reported ' + result.error );
            return updater.onError( result );
        } else {
            updater.pollCount -= 1;
            updater._onSuccess( result );
            updater.timer = window.setTimeout(updater.poll, updater.pollInterval);
        }
    };
    updater._onError = function( response ) {
        for (var i=0;i<updater.failureCallbacks.length;i++) {
            try {
                updater.failureCallbacks[i]( response );
            } catch( e ) {
                console.log( "Failure during failure callback "+updater.failureCallbacks[i]);
            }
        }
    };
    updater.onError = function( response, statustext ) {
        updater.pollCount -= 1;
        if (updater.request === response) {
            /* we are the current live response */
            updater.request = null;
        }
        try {
            if (response.abort) {
                response.abort();
            }
        } catch (e) {
            /* ARGH IE sucks! */
        }
        if (statustext == 'timeout') {
            console.log("Poll timeout, re-requesting");
            return updater.poll();
        } else {
            updater.errorSleepTime *= 2;
            if (updater.errorSleepTime > 32000) {
                updater.errorSleepTime = 16000;
            }
            /* there's no other error-timeout waiting */
            console.log("Poll error; sleeping for", updater.errorSleepTime, "ms");
            updater.timer = window.setTimeout(updater.poll, updater.errorSleepTime);
        }
        return updater._onError( response );
    };
    updater.stop = function() {
        updater.abortCurrent();
        updater.live = false;
    };
    updater.restart = function() {
        updater.abortCurrent();
        updater.live = true;
        return updater.poll();
    };
    updater._streamCallback = function( xhr ) {
        /* creates stream-callback function for xhr */
        /* TODO: use this for *all* processing, maybe split out the jquery-specific bits of the implementation */
        var bytesRead = 0;
        var callback = function( ) {
            console.log( 'Ready state '+xhr.readyState );
            if (xhr.readyState == 3 || xhr.readyState == 4) {
                /* interactive or finished */
                var text;
                console.log( 'Raw text '+ xhr.responseText );
                if (xhr.readyState ==4) {
                    /* we're done, there may not be a split-marker coming */
                    if (bytesRead < xhr.responseText.length) {
                        text = [ xhr.responseText.substring( bytesRead ) ];
                    } else {
                        text = [];
                    }
                } else {
                    text = xhr.responseText.substring( bytesRead ).split( updater.MESSAGE_SPLIT_TEXT );
                }
                if (text && text.length) {
                    // skip the last element, if we haven't got a split after it it isn't considered fully-delivered...
                    text.length = text.length -1;
                    text.forEach( function(t) {
                        if (!t.length) {
                            return ;
                        }
                        try {
                            /* TODO: use a json parser! */
                            /* TODO: add a divider to the JSON responses to make parsing safer/easier */
                            console.log( 'Got partial response '+ t );
                            var json;
                            json = JSON.parse( t );
                            if (json.error) {
                                updater._onError( json);
                            } else {
                                updater._onSuccess( json );
                            }
                            console.log( 'Read '+t.length+ ' bytes' );
                            bytesRead = bytesRead + t.length + updater.MESSAGE_SPLIT_TEXT.length;
                        } catch (e) {
                            /* TODO: only consider format errors , get console.log working in exception handlers */
                            console.log( 'Failure '+ e );
                        }
                    });
                    if (xhr.readyState == 3 && xhr.responseText.length > updater.MAX_STREAM_DATA) {
                        /* we've had too much data down this channel, force a memory dump... */
                        /* TODO: make sure we get an onError here! */
                        xhr.abort();
                        console.log( 'Restarting XHR channel to dump memory load' );
                        updater.pollCount -= 1;
                        updater.poll();
                    }
                }
            }
        };
        return callback;
    };
    updater.poll = function( ignoredups ) {
        if (updater.live) {
            if (updater.pollCount > 0) {
                if (!ignoredups) {
                    console.log( 'longPoll already running '+updater.pollCount );
                }
                return false;
            }
            updater.pollCount += 1;
            var args = {};
            var callback = updater.onSuccess;
            var timeout = updater.errorSleepTime * 10;
            if (updater.pollInterval) {
                args.noWait = true;
                callback = updater.onSuccessNoWait;
                timeout = updater.errorSleepTime;
            }
            if (updater.stream) {
                timeout = null;
                args.stream = true;
            }
            $.extend( args, updater.baseArguments );
            if (updater.calcArguments) {
                $.extend( args, updater.calcArguments( updater ));
            }
            console.log( 'polling'+ JSON.stringify( args ) );
            var xhr = $.ajax({
                url: updater.url,
                type: updater.method,
                dataType: "text",
                data: $.param(args),
                success: updater.onSuccess,
                error: updater.onError,
                cache: false,
                /* allow the request to "hang" for a good long while IFF we're doing long-polling */
                timeout: timeout
            });
            /* set xhr request handler for partial responses... */
            updater.request = xhr;
            if (updater.stream) {
                xhr.onreadystatechange = updater._streamCallback( xhr );
            }
        } else {
            console.log( 'Exiting longPoll '+updater );
            return false;
        }
        return true;
    };
    return updater;
};

$.torChannels = function( config, updater ) {
    /* Create a simplified channel-watching (long-polling) subscription service

    if no updater is passed, the update is created by calling longPoll( config )

    returns Channels object.  This is intended to work with a server-side publish/subscribe
    service to provide a simplistic implementation of channel subscriptions.  This service
    should be a long-polling asynchronous handler hooked up e.g. to the channels
    implementation provided.  The channel ids subscribed and their related timestamps
    will be passed as the body of the JSON post.

    Attributes of the Channels object:

        updater -- the longPoll object we are using

        last_ts -- the last timestamp we received from the server

        channels -- { channel_id : subscription } mapping

    Methods of the Channels object:

        subscribe( channel_id, query, callback, ts ) -> subinfo

            subscribe this callback to messages on the given channel, returns an object
            which can be used to unsubscribe later.  ts is the timestamp from which to
            request messages, normally you set this to the timestamp of your out-of-band
            query which retrieved initial state.

            query -- dictionary describing the query to use to determine "interest" on the
                server-side.  Normally a series of ranged key-values, such as:

                    ts: [ first_ts, null ]

                which would signal to return any message with a key of 'ts' greater than
                ts, with no upper condition...

        unsubscribe( subinfo )

            unsubscribe a previously-subscribed callback from the channel

    */
    if (! updater) {
        updater = $.longPoll( config );
    }
    var subscriptions = {
        updater: updater,
        last_ts: 0.0,
        channels: []
    };
    $.extend( subscriptions, config );

    subscriptions._updateArgs = function() {
        var myChannels = {};
        subscriptions.channels.forEach( function( sub ) {
            myChannels[ sub.key ] = sub.queries;
        } );
        return {
            channels: JSON.stringify(  myChannels ),
            cursor: subscriptions.last_ts
        };
    };
    if (subscriptions.updater.calcArguments) {
        var old_calcArgs = subscriptions.updater.calcArguments;
        subscriptions.updater.calcArguments = function( updater ) {
            var base = old_calcArgs( updater );
            $.extend( base, subscriptions._updateArgs());
            return base;
        };
    } else {
        subscriptions.updater.calcArguments = subscriptions._updateArgs;
    }
    subscriptions.get_sub = function( channel ) {
        /* get any existing sub for channel */
        for (var i=0;i<subscriptions.channels.length;i++) {
            var sub = subscriptions.channels[i];
            if (sub.key == channel) {
                return sub;
            }
        }
        return null;
    };
    subscriptions.subscription = function( channel, nocreate ) {
        /* create a new subscription object */
        var sub = subscriptions.get_sub( channel );
        if (nocreate || sub) {
            return sub;
        }
        sub = {
            key: channel,
            queries: [],
            callbacks: []
        };
        sub.process = function( message ) {
            /* Process incoming message in this channel */
            if (message.channel != sub.key) {
                console.log( 'Subscription got a message that does not match key???' );
                return false;
            }
            sub.callbacks.forEach( function(cb) {
                try {
                    cb( message, sub );
                } catch (e) {
                    console.log(
                        "Failure in channel callback "+ e + ': '+cb
                    );
                }
            } );
        };
        sub.addCallback = function( callback,query ) {
            /* add the callback and return object to unsubscribe later */
            sub.callbacks.push( callback );
            sub.queries.push( query );
            return {
                channel: channel,
                query: query,
                callback: callback
            };
        };
        sub.removeCallback = function( callback ) {
            for (var i=sub.callbacks.length-1;i>-1;i--) {
                if (sub.callbacks[i] === callback) {
                    sub.callbacks.splice( i,1 );
                    sub.queries.splice( i, 1 );
                }
            }
        };
        subscriptions.channels.push( sub );
        return sub;
    };
    subscriptions.restart = function( ) {
        subscriptions.updater.abortCurrent();
        subscriptions.updater.poll(true); /* ignore duplicate registrations for this call */
    };
    subscriptions.subscribe = function( channel, query, callback, ts, no_abort ) {
        var sub = subscriptions.subscription( channel );
        sub.ts = ts;
        var subinfo = sub.addCallback( callback, query );
        if (!no_abort) {
            subscriptions.restart();
        }
        return subinfo;
    };
    subscriptions.unsubscribe = function( subinfo ) {
        var sub = subscriptions.subscription( subinfo.channel, true );
        if (sub) {
            sub.removeCallback( subinfo.callback );
            if (!sub.callbacks.length) {
                for (var i=subscriptions.channels.length-1;i>-1;i--) {
                    if (subscriptions.channels[i].key == sub.key) {
                        subscriptions.channels.splice( i,1 );
                    }
                }
            }
            /* TODO: if there are *no* channels, stop the long-poller entirely */
            subscriptions.updater.abortCurrent();
            subscriptions.updater.poll(true); /* ignore duplicate registrations for this call */
            return subinfo;
        }
    };
    subscriptions._onMessages = function( message ) {
        /* handle a top-level message */
        if (!message.messages) {
            console.log( 'Received non-subscription message ' +message );
            return;
        }
        if (message.ts && message.ts > subscriptions.last_ts) {
            subscriptions.last_ts = message.ts;
        }
        message.messages.forEach( function( msg ) {
            var sub = subscriptions.subscription( msg.channel, true );
            if (sub) {
                sub.process( msg );
            } else {
                console.log( 'Receive unwanted subscription message for channel ignoring '+msg.channel );
            }
        } );
    };
    updater.callbacks.push( subscriptions._onMessages );
    return subscriptions;
};
