"""Handler mix-ins for longPoll implementation

TODO: Add an MSIE-compatible streaming transport (instead of long-polling).
TODO: Test on iPhone browser.
TODO: Test on IE 6

TODO: Add "last drop timestamp" to channels so that they can report a
    client request that is stale (is older than the trim-date which dropped
    messages from the channel)

TODO: Make it possible to specify per-channel ts values (again), so that
    we can catch up on channel X without having to code to ignore/transfer
    all of channel Y's already-received messages

TODO: "Persistent Channel" implementation???  If we have a state mechanism
    then we're essentially storing data... which suggests that we'll want
    that data back some day...

TODO: Add a multi-process transport to allow for splitting channels among
    multiple machines/processes (scale horizontally for single channel)
TODO: Provide sample code for load-balancing channels to different
    machines (scale horizontally for multiple channels)

TODO: Packaging and documentation
TODO: Add example application/code (or wait until I release BackChannel).
TODO: Twisted implementation for the back-end.
"""
import time,logging,re
log = logging.getLogger( 'longpoll' )
import tornado.web
from torchannels import channels
from torchannels import basechannels
from gettext import gettext as _

class ChannelMixin( basechannels.BaseChannels ):

    @tornado.web.authenticated
    @tornado.web.asynchronous
    def post(self):
        """Post request for new messages

        channels -- channels to which to subscribe/listen

        noWait -- if present, return immediately (no long-polling)

        stream -- if present, attempt to do a streaming response

        ts -- timestamp of the last request (from the server)

        chan_ts -- if present, a json-encoded dictionary with the
            channel timestamps

        returns JSON of either:

            {'error': 'explanation','ts':timestamp}
            {'messages': [{}...],'ts':timestamp}
        """
        subscriptions = self.request.arguments.get('channels', [])
        if isinstance( subscriptions, (str,unicode)):
            subscriptions = [subscriptions]
        if not subscriptions:
            self.write( {
                'error': _('No subscriptions specified'),
                'ts': time.time(),
            } )
            return self.finish()
        timestamp = float(self.get_argument('cursor', 0.0))
        log.info(
            "Polling subscriptions: %s, %s",
            timestamp,
            subscriptions,
        )
        # check that we got valid channel names
        for sub in subscriptions:
            if not self.can_access( sub ):
                self.write({
                    'error': _('Invalid channel name %(sub)r')%dict(sub=sub),
                })
                return self.finish()
        self.stream = stream = self.get_argument( 'stream', False )
        log.info( 'Streaming poll? %s', stream )
        pending = self.updates( subscriptions, timestamp )
        if pending:
            self.write( dict(
                messages = pending,
                ts = time.time(),
            ) )
            if not stream:
                log.info( 'Long-polling request finishing at setup' )
                self.finish()
                return
            else:
                log.info( 'Streaming request waiting after initial write' )
                self.write( self.MESSAGE_SPLIT_TEXT )
                self.flush()
        else:
            if self.get_argument( 'quickPoll', False ):
                log.info( 'Quick poll returning nothing' )
                self.write( { 'ok': True, 'messages': [] } )
                return self.finish()
            else:
                log.info( 'Long poll/streaming waiting...' )
        return self.subscribe( subscriptions, self.on_message )

    @property
    def invalid( self ):
        """Are we already marked invalid (closed or finished)?"""
        if self.request.connection.stream.closed() or self._finished:
            return True
        return False

    def on_message( self, message ):
        """Process message on channel"""
        if self.invalid:
            return None
        log.info( 'Sending delayed post response: %s', message )
        try:
            self.write( dict(
                messages=[message],
                ts = message['ts'],
            ) )
        except IOError, err:
            return None
        if not self.stream:
            log.info( 'Long-polling request finishing on update' )
            try:
                return self.finish()
            except IOError, err:
                return None
        else:
            log.info( 'Streaming request continuing on update' )
            try:
                self.write( self.MESSAGE_SPLIT_TEXT )
                self.flush()
            except IOError, err:
                assert self.invalid, '''Unable to write, but we aren't invalid???'''
